# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import logging
from functools import partial
from typing import Any, Mapping, Optional

from PySide6.QtCore import Signal
from PySide6.QtWidgets import QVBoxLayout, QWidget

from acconeer.exptool import a121
from acconeer.exptool.a121._core import Criticality

from . import pidgets
from .range_help_view import RangeHelpView
from .types import PidgetFactoryMapping
from .utils import VerticalGroupBox


log = logging.getLogger(__name__)


class SensorConfigEditor(QWidget):
    sig_update = Signal(object)

    _sensor_config: Optional[a121.SensorConfig]

    _all_pidgets: list[pidgets.ParameterWidget]

    SPACING = 15
    IDLE_STATE_LABEL_MAP = {
        a121.IdleState.READY: "Ready",
        a121.IdleState.SLEEP: "Sleep",
        a121.IdleState.DEEP_SLEEP: "Deep sleep",
    }
    PROFILE_LABEL_MAP = {
        a121.Profile.PROFILE_1: "1 (shortest)",
        a121.Profile.PROFILE_2: "2",
        a121.Profile.PROFILE_3: "3",
        a121.Profile.PROFILE_4: "4",
        a121.Profile.PROFILE_5: "5 (longest)",
    }
    PRF_LABEL_MAP = {
        a121.PRF.PRF_19_5_MHz: "19.5 MHz",
        a121.PRF.PRF_13_0_MHz: "13.0 MHz",
        a121.PRF.PRF_8_7_MHz: "8.7 MHz",
        a121.PRF.PRF_6_5_MHz: "6.5 MHz",
    }
    SENSOR_CONFIG_FACTORIES: PidgetFactoryMapping = {
        "sweeps_per_frame": pidgets.IntParameterWidgetFactory(
            name_label_text="Sweeps per frame:",
            limits=(1, 4095),
        ),
        "sweep_rate": pidgets.OptionalFloatParameterWidgetFactory(
            name_label_text="Sweep rate:",
            limits=(1, 1e6),
            decimals=0,
            init_set_value=1000.0,
            suffix="Hz",
            checkbox_label_text="Limit",
        ),
        "frame_rate": pidgets.OptionalFloatParameterWidgetFactory(
            name_label_text="Frame rate:",
            limits=(0.1, 1e4),
            decimals=1,
            init_set_value=10.0,
            suffix="Hz",
            checkbox_label_text="Limit",
        ),
        "inter_sweep_idle_state": pidgets.EnumParameterWidgetFactory(
            enum_type=a121.IdleState,
            name_label_text="Inter sweep idle state:",
            label_mapping=IDLE_STATE_LABEL_MAP,
        ),
        "inter_frame_idle_state": pidgets.EnumParameterWidgetFactory(
            enum_type=a121.IdleState,
            name_label_text="Inter frame idle state:",
            label_mapping=IDLE_STATE_LABEL_MAP,
        ),
        "continuous_sweep_mode": pidgets.CheckboxParameterWidgetFactory(
            name_label_text="Continuous sweep mode",
        ),
    }
    SUBSWEEP_CONFIG_FACTORIES: PidgetFactoryMapping = {
        "start_point": pidgets.IntParameterWidgetFactory(
            name_label_text="Start point:",
        ),
        "num_points": pidgets.IntParameterWidgetFactory(
            name_label_text="Number of points:",
            limits=(1, 4095),
        ),
        "step_length": pidgets.IntParameterWidgetFactory(
            name_label_text="Step length:",
            limits=(1, None),
        ),
        "hwaas": pidgets.IntParameterWidgetFactory(
            name_label_text="HWAAS:",
            limits=(1, 511),
        ),
        "receiver_gain": pidgets.IntParameterWidgetFactory(
            name_label_text="Receiver gain:",
            limits=(0, 23),
        ),
        "profile": pidgets.EnumParameterWidgetFactory(
            enum_type=a121.Profile,
            name_label_text="Profile:",
            label_mapping=PROFILE_LABEL_MAP,
        ),
        "prf": pidgets.EnumParameterWidgetFactory(
            enum_type=a121.PRF,
            name_label_text="PRF:",
            label_mapping=PRF_LABEL_MAP,
        ),
        "enable_tx": pidgets.CheckboxParameterWidgetFactory(
            name_label_text="Enable transmitter",
        ),
        "enable_loopback": pidgets.CheckboxParameterWidgetFactory(
            name_label_text="Enable loopback",
        ),
        "phase_enhancement": pidgets.CheckboxParameterWidgetFactory(
            name_label_text="Phase enhancement",
        ),
    }

    def __init__(self, parent: Optional[QWidget] = None) -> None:
        super().__init__(parent=parent)

        self._all_pidgets = []

        self._sensor_config = None

        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)

        # Sensor pidgets

        self.sensor_group_box = VerticalGroupBox("Sensor parameters", parent=self)
        self.sensor_group_box.layout().setSpacing(self.SPACING)
        self.layout().addWidget(self.sensor_group_box)

        self._sensor_config_pidgets: Mapping[str, pidgets.ParameterWidget] = {}
        for aspect, factory in self.SENSOR_CONFIG_FACTORIES.items():
            pidget = factory.create(self.sensor_group_box)
            self.sensor_group_box.layout().addWidget(pidget)

            pidget.sig_parameter_changed.connect(
                partial(self._update_sensor_config_aspect, aspect)
            )

            self._all_pidgets.append(pidget)
            self._sensor_config_pidgets[aspect] = pidget

        # Subsweeps pidgets

        self.subsweep_group_box = VerticalGroupBox(
            "Subsweep-specific parameters", parent=self.sensor_group_box
        )
        self.subsweep_group_box.layout().setSpacing(self.SPACING)
        self.sensor_group_box.layout().addWidget(self.subsweep_group_box)

        self.range_help_view = RangeHelpView(self.subsweep_group_box)
        self.subsweep_group_box.layout().addWidget(self.range_help_view)

        self._subsweep_config_pidgets: Mapping[str, pidgets.ParameterWidget] = {}
        for aspect, factory in self.SUBSWEEP_CONFIG_FACTORIES.items():
            pidget = factory.create(self.subsweep_group_box)
            self.subsweep_group_box.layout().addWidget(pidget)

            pidget.sig_parameter_changed.connect(
                partial(self._update_subsweep_config_aspect, aspect)
            )

            self._all_pidgets.append(pidget)
            self._subsweep_config_pidgets[aspect] = pidget

    def _update_ui(self) -> None:
        if self._sensor_config is None:
            log.debug("could not update ui as SensorConfig is None")
            return

        self._sensor_config_pidgets["sweeps_per_frame"].set_parameter(
            self._sensor_config.sweeps_per_frame
        )
        self._sensor_config_pidgets["sweep_rate"].set_parameter(self._sensor_config.sweep_rate)
        self._sensor_config_pidgets["frame_rate"].set_parameter(self._sensor_config.frame_rate)
        self._sensor_config_pidgets["continuous_sweep_mode"].set_parameter(
            self._sensor_config.continuous_sweep_mode
        )
        self._sensor_config_pidgets["inter_frame_idle_state"].set_parameter(
            self._sensor_config.inter_frame_idle_state
        )
        self._sensor_config_pidgets["inter_sweep_idle_state"].set_parameter(
            self._sensor_config.inter_sweep_idle_state
        )

        subsweep_config = self._sensor_config.subsweep
        self._subsweep_config_pidgets["start_point"].set_parameter(subsweep_config.start_point)
        self._subsweep_config_pidgets["num_points"].set_parameter(subsweep_config.num_points)
        self._subsweep_config_pidgets["step_length"].set_parameter(subsweep_config.step_length)
        self._subsweep_config_pidgets["profile"].set_parameter(subsweep_config.profile)
        self._subsweep_config_pidgets["hwaas"].set_parameter(subsweep_config.hwaas)
        self._subsweep_config_pidgets["receiver_gain"].set_parameter(subsweep_config.receiver_gain)
        self._subsweep_config_pidgets["enable_tx"].set_parameter(subsweep_config.enable_tx)
        self._subsweep_config_pidgets["enable_loopback"].set_parameter(
            subsweep_config.enable_loopback
        )
        self._subsweep_config_pidgets["phase_enhancement"].set_parameter(
            subsweep_config.phase_enhancement
        )
        self._subsweep_config_pidgets["prf"].set_parameter(subsweep_config.prf)

    def set_data(self, sensor_config: Optional[a121.SensorConfig]) -> None:
        self._sensor_config = sensor_config
        self.range_help_view.update(sensor_config.subsweep if sensor_config else None)

    def sync(self) -> None:
        self._update_ui()

    def _broadcast(self) -> None:
        self.sig_update.emit(self._sensor_config)

    def _handle_validation_results(self, results: list[a121.ValidationResult]) -> None:
        if results == []:
            for pidget in self._all_pidgets:
                pidget.set_note_text("")
        else:
            for result in results:
                self._handle_validation_result(result)

    def _handle_validation_result(self, result: a121.ValidationResult) -> None:
        if result.aspect is None or self._sensor_config is None:
            return
        if result.source is self._sensor_config:
            pidget_map = self._sensor_config_pidgets
        elif result.source is self._sensor_config.subsweep:
            pidget_map = self._subsweep_config_pidgets
        else:
            return

        pidget_map[result.aspect].set_note_text(result.message, result.criticality)

    def _update_sensor_config_aspect(self, aspect: str, value: Any) -> None:
        if self._sensor_config is None:
            raise TypeError("SensorConfig is None")

        try:
            setattr(self._sensor_config, aspect, value)
        except Exception as e:
            self._sensor_config_pidgets[aspect].set_note_text(e.args[0], Criticality.ERROR)
        else:
            self._handle_validation_results(self._sensor_config._collect_validation_results())

        self._broadcast()

    def _update_subsweep_config_aspect(self, aspect: str, value: Any) -> None:
        if self._sensor_config is None:
            raise TypeError("SensorConfig is None")

        try:
            setattr(self._sensor_config.subsweep, aspect, value)
        except Exception as e:
            self._subsweep_config_pidgets[aspect].set_note_text(e.args[0], Criticality.ERROR)
        else:
            self._handle_validation_results(self._sensor_config._collect_validation_results())

        self._broadcast()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import numpy as np

import pyqtgraph as pg

import acconeer.exptool as et
from acconeer.exptool import a121
from acconeer.exptool.a121.algo.distance import (
    Aggregator,
    AggregatorConfig,
    AggregatorResult,
    ProcessorConfig,
    ProcessorMode,
    ProcessorSpec,
    ThresholdMethod,
)


def main():
    args = a121.ExampleArgumentParser().parse_args()
    et.utils.config_logging(args)

    client = a121.Client(**a121.get_client_args(args))
    client.connect()

    sensor_id = 1

    sensor_config_1 = a121.SensorConfig(
        subsweeps=[
            a121.SubsweepConfig(
                start_point=50,
                step_length=4,
                num_points=25,
                profile=a121.Profile.PROFILE_1,
                hwaas=10,
                phase_enhancement=True,
            ),
            a121.SubsweepConfig(
                start_point=150,
                step_length=4,
                num_points=25,
                profile=a121.Profile.PROFILE_1,
                hwaas=20,
                phase_enhancement=True,
            ),
            a121.SubsweepConfig(
                start_point=200,
                step_length=4,
                num_points=50,
                profile=a121.Profile.PROFILE_2,
                hwaas=20,
                phase_enhancement=True,
            ),
        ],
        sweeps_per_frame=1,
    )

    sensor_config_2 = a121.SensorConfig(
        profile=a121.Profile.PROFILE_2,
        start_point=350,
        step_length=4,
        num_points=25,
        phase_enhancement=True,
        sweeps_per_frame=10,
    )

    session_config = a121.SessionConfig(
        [{sensor_id: sensor_config_1}, {sensor_id: sensor_config_2}], extended=True
    )
    extended_metadata = client.setup_session(session_config)

    aggregator_config = AggregatorConfig()

    processor_config = ProcessorConfig(
        processor_mode=ProcessorMode.DISTANCE_ESTIMATION,
        threshold_method=ThresholdMethod.FIXED,
        fixed_threshold_value=500,
    )

    processor_specs = [
        ProcessorSpec(
            processor_config=processor_config,
            group_index=0,
            sensor_id=sensor_id,
            subsweep_indexes=[0, 1],
        ),
        ProcessorSpec(
            processor_config=processor_config,
            group_index=0,
            sensor_id=sensor_id,
            subsweep_indexes=[2],
        ),
        ProcessorSpec(
            processor_config=processor_config,
            group_index=1,
            sensor_id=sensor_id,
            subsweep_indexes=[0],
        ),
    ]

    aggregator = Aggregator(
        session_config=session_config,
        extended_metadata=extended_metadata,
        config=aggregator_config,
        specs=processor_specs,
    )

    pg_updater = PGUpdater(num_sweeps=len(processor_specs))
    pg_process = et.PGProcess(pg_updater)
    pg_process.start()

    client.start_session()

    interrupt_handler = et.utils.ExampleInterruptHandler()
    print("Press Ctrl-C to end session")

    while not interrupt_handler.got_signal:
        extended_result = client.get_next()
        processed_data = aggregator.process(extended_result)
        try:
            pg_process.put_data(processed_data)
        except et.PGProccessDiedException:
            break

    print("Disconnecting...")
    client.disconnect()


class PGUpdater:
    def __init__(self, num_sweeps: int):
        self.num_sweeps = num_sweeps

    def setup(self, win):
        self.sweep_plot = win.addPlot(row=0, col=0)
        self.sweep_plot.setMenuEnabled(False)
        self.sweep_plot.showGrid(x=True, y=True)
        self.sweep_plot.addLegend()
        self.sweep_plot.setLabel("left", "Amplitude")
        self.sweep_plot.addItem(pg.PlotDataItem())

        pen = et.utils.pg_pen_cycler(0)
        brush = et.utils.pg_brush_cycler(0)
        symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
        feat_kw = dict(pen=pen, **symbol_kw)
        self.sweep_curves = [self.sweep_plot.plot(**feat_kw) for _ in range(self.num_sweeps)]

        pen = et.utils.pg_pen_cycler(1)
        brush = et.utils.pg_brush_cycler(1)
        symbol_kw = dict(symbol="o", symbolSize=1, symbolBrush=brush, symbolPen="k")
        feat_kw = dict(pen=pen, **symbol_kw)
        self.threshold_curves = [self.sweep_plot.plot(**feat_kw) for _ in range(self.num_sweeps)]

    def update(self, result: AggregatorResult):

        for idx, processor_result in enumerate(result.processor_results):
            threshold = processor_result.extra_result.used_threshold
            valid_threshold_idx = np.where(~np.isnan(threshold))[0]
            threshold = threshold[valid_threshold_idx]
            self.sweep_curves[idx].setData(
                processor_result.extra_result.distances_m, processor_result.extra_result.abs_sweep
            )
            self.threshold_curves[idx].setData(
                processor_result.extra_result.distances_m[valid_threshold_idx], threshold
            )


if __name__ == "__main__":
    main()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import Mapping

from .pidgets import ParameterWidgetFactory


PidgetFactoryMapping = Mapping[str, ParameterWidgetFactory]



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from ._aggregator import (
    Aggregator,
    AggregatorConfig,
    AggregatorResult,
    PeakSortingMethod,
    ProcessorSpec,
)
from ._detector import Detector, DetectorConfig, DetectorResult
from ._processors import (
    MeasurementType,
    Processor,
    ProcessorConfig,
    ProcessorContext,
    ProcessorMode,
    ProcessorResult,
    ThresholdMethod,
)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import numpy as np
from numpy import cos, pi, sqrt, square
from scipy.special import binom

import acconeer.exptool as et


def get_sensor_config():
    config = et.a111.SparseServiceConfig()
    config.profile = et.a111.SparseServiceConfig.Profile.PROFILE_3
    config.sampling_mode = et.a111.SparseServiceConfig.SamplingMode.B
    config.range_interval = [0.3, 1.3]
    config.update_rate = 80
    config.sweeps_per_frame = 32
    config.hw_accelerated_average_samples = 60
    return config


class ProcessingConfiguration(et.configbase.ProcessingConfig):
    VERSION = 6

    detection_threshold = et.configbase.FloatParameter(
        label="Detection threshold",
        default_value=1.5,
        limits=(0, 5),
        updateable=True,
        order=0,
        help='Level at which the detector output is considered as "present".',
    )

    inter_frame_fast_cutoff = et.configbase.FloatParameter(
        label="Inter fast cutoff freq.",
        unit="Hz",
        default_value=20.0,
        limits=(1, 100),
        logscale=True,
        updateable=True,
        order=10,
        help=(
            "Cutoff frequency of the low pass filter for the fast filtered sweep mean."
            " No filtering is applied if the cutoff is set over half the frame rate"
            " (Nyquist limit)."
        ),
    )

    inter_frame_slow_cutoff = et.configbase.FloatParameter(
        label="Inter slow cutoff freq.",
        unit="Hz",
        default_value=0.2,
        limits=(0.01, 1),
        logscale=True,
        updateable=True,
        order=20,
        help="Cutoff frequency of the low pass filter for the slow filtered sweep mean.",
    )

    inter_frame_deviation_time_const = et.configbase.FloatParameter(
        label="Inter deviation time const.",
        unit="s",
        default_value=0.5,
        limits=(0.01, 30),
        logscale=True,
        updateable=True,
        order=30,
        help=(
            "Time constant of the low pass filter for the (inter-frame) deviation between"
            " fast and slow."
        ),
    )

    intra_frame_time_const = et.configbase.FloatParameter(
        label="Intra time const.",
        unit="s",
        default_value=0.15,
        limits=(0, 0.5),
        updateable=True,
        order=40,
        help="Time constant for the intra frame part.",
    )

    intra_frame_weight = et.configbase.FloatParameter(
        label="Intra weight",
        default_value=0.6,
        limits=(0, 1),
        updateable=True,
        order=50,
        help=(
            "The weight of the intra-frame part in the final output. A value of 1 corresponds"
            " to only using the intra-frame part and a value of 0 corresponds to only using"
            " the inter-frame part."
        ),
    )

    output_time_const = et.configbase.FloatParameter(
        label="Output time const.",
        unit="s",
        default_value=0.5,
        limits=(0.01, 30),
        logscale=True,
        updateable=True,
        order=60,
        help="Time constant of the low pass filter for the detector output.",
    )

    num_removed_pc = et.configbase.IntParameter(
        label="PCA based noise reduction",
        default_value=0,
        limits=(0, 2),
        updateable=False,
        order=70,
        help=(
            "Sets the number of principal components removed in the PCA based noise reduction."
            " Filters out static reflections."
            " Setting to 0 (default) disables the PCA based noise reduction completely."
        ),
    )

    show_data = et.configbase.BoolParameter(
        label="Show data scatter plot",
        default_value=True,
        updateable=True,
        order=100,
        help=(
            "Show the plot of the current data frame along with the fast and slow filtered"
            " mean sweep (used in the inter-frame part)."
        ),
    )

    show_noise = et.configbase.BoolParameter(
        label="Show noise",
        default_value=False,
        updateable=True,
        order=110,
        help="Show the noise estimation plot.",
        category=et.configbase.Category.ADVANCED,
    )

    show_depthwise_output = et.configbase.BoolParameter(
        label="Show depthwise presence",
        default_value=True,
        updateable=True,
        order=120,
        help="Show the depthwise presence output plot.",
    )

    show_sectors = et.configbase.BoolParameter(
        label="Show distance sectors",
        default_value=False,
        updateable=True,
        order=130,
    )

    history_plot_ceiling = et.configbase.FloatParameter(
        label="Presence score plot ceiling",
        default_value=10.0,
        decimals=1,
        limits=(1, 100),
        logscale=True,
        updateable=True,
        optional=True,
        optional_label="Fixed",
        order=190,
        help="The highest presence score that will be plotted.",
        category=et.configbase.Category.ADVANCED,
    )

    history_length_s = et.configbase.FloatParameter(
        label="History length",
        unit="s",
        default_value=5,
        limits=(1, 20),
        decimals=0,
        order=200,
        category=et.configbase.Category.ADVANCED,
    )

    def check_sensor_config(self, conf):
        alerts = {
            "processing": [],
            "sensor": [],
        }
        if conf.update_rate is None:
            alerts["sensor"].append(et.configbase.Error("update_rate", "Must be set"))

        if not conf.sweeps_per_frame > 3:
            alerts["sensor"].append(et.configbase.Error("sweeps_per_frame", "Must be > 3"))

        return alerts


class Processor:
    # lp(f): low pass (filtered)
    # cut: cutoff frequency [Hz]
    # tc: time constant [s]
    # sf: smoothing factor [dimensionless]

    def __init__(self, sensor_config, processing_config, session_info, calibration=None):
        self.sweeps_per_frame = sensor_config.sweeps_per_frame
        self.depths = et.a111.get_range_depths(sensor_config, session_info)
        self.num_depths = self.depths.size
        self.f = sensor_config.update_rate
        self.num_removed_pc = processing_config.num_removed_pc

        # Fixed parameters
        self.noise_est_diff_order = 3
        self.depth_filter_length = 3
        noise_tc = 1.0

        assert sensor_config.update_rate is not None
        assert self.sweeps_per_frame > self.noise_est_diff_order

        self.noise_sf = self.tc_to_sf(noise_tc, self.f)

        nd = self.noise_est_diff_order
        self.noise_norm_factor = np.sqrt(np.sum(np.square(binom(nd, np.arange(nd + 1)))))

        self.fast_lp_mean_sweep = np.zeros(self.num_depths)
        self.slow_lp_mean_sweep = np.zeros(self.num_depths)
        self.lp_inter_dev = np.zeros(self.num_depths)
        self.lp_intra_dev = np.zeros(self.num_depths)
        self.lp_noise = np.zeros(self.num_depths)

        if self.num_removed_pc == 0:
            self.noise_base = None
        else:
            self.noise_base = np.ones([self.num_removed_pc, self.num_depths])
            if self.num_removed_pc == 2:
                self.noise_base[1, ::2] = -1

            self.noise_base = self.normalize_noise_base(self.noise_base)

        self.presence_score = 0
        self.presence_distance_index = 0
        self.presence_distance = 0

        self.presence_history = np.zeros(int(round(self.f * processing_config.history_length_s)))
        self.update_index = 0

        self.update_processing_config(processing_config)

    def update_processing_config(self, processing_config):
        self.threshold = processing_config.detection_threshold
        self.intra_weight = processing_config.intra_frame_weight
        self.inter_weight = 1.0 - self.intra_weight

        self.fast_sf = self.cutoff_to_sf(processing_config.inter_frame_fast_cutoff, self.f)
        self.slow_sf = self.cutoff_to_sf(processing_config.inter_frame_slow_cutoff, self.f)
        self.inter_dev_sf = self.tc_to_sf(
            processing_config.inter_frame_deviation_time_const, self.f
        )
        self.intra_sf = self.tc_to_sf(processing_config.intra_frame_time_const, self.f)
        self.output_sf = self.tc_to_sf(processing_config.output_time_const, self.f)

    def cutoff_to_sf(self, fc, fs):  # cutoff frequency to smoothing factor conversion
        if fc > 0.5 * fs:
            return 0.0

        cos_w = cos(2.0 * pi * (fc / fs))
        return 2.0 - cos_w - sqrt(square(cos_w) - 4.0 * cos_w + 3.0)

    def tc_to_sf(self, tc, fs):  # time constant to smoothing factor conversion
        if tc <= 0.0:
            return 0.0

        return np.exp(-1.0 / (tc * fs))

    def dynamic_sf(self, static_sf):
        return min(static_sf, 1.0 - 1.0 / (1.0 + self.update_index))

    def abs_dev(self, a, axis=None, ddof=0, subtract_mean=True):
        if subtract_mean:
            a = a - a.mean(axis=axis, keepdims=True)

        if axis is None:
            n = a.size
        else:
            n = a.shape[axis]

        assert ddof >= 0
        assert n > ddof

        return np.mean(np.abs(a), axis=axis) * sqrt(n / (n - ddof))

    def depth_filter(self, a):
        b = np.ones(self.depth_filter_length) / self.depth_filter_length

        if a.size >= b.size:
            return np.correlate(a, b, mode="same")
        else:
            pad_width = int(np.ceil((b.size - a.size) / 2))
            a = np.pad(a, pad_width, "constant")
            return np.correlate(a, b, mode="same")[pad_width:-pad_width]

    @staticmethod
    def normalize_noise_base(noise_base):
        norm = np.sqrt(np.sum(np.square(noise_base), axis=1, keepdims=True))
        return noise_base / norm

    def process(self, data, data_info):
        frame = data

        # Noise estimation

        nd = self.noise_est_diff_order

        noise_diff = np.diff(frame, n=nd, axis=0)
        noise = self.abs_dev(noise_diff, axis=0, subtract_mean=False)
        noise /= self.noise_norm_factor
        sf = self.dynamic_sf(self.noise_sf)
        self.lp_noise = sf * self.lp_noise + (1.0 - sf) * noise

        # Intra-frame part

        mean_sweep = frame.mean(axis=0)
        frame_diff = frame - mean_sweep

        if self.num_removed_pc > 0:
            noise_coeffs = frame_diff @ self.noise_base.transpose()
            frame_diff = np.abs(frame_diff) - np.abs(noise_coeffs @ self.noise_base)
            frame_diff = np.maximum(frame_diff, 0)

        sweep_dev = self.abs_dev(frame_diff, axis=0, ddof=1, subtract_mean=False)

        sf = self.dynamic_sf(self.intra_sf)
        self.lp_intra_dev = sf * self.lp_intra_dev + (1.0 - sf) * sweep_dev

        norm_lp_intra_dev = np.divide(
            self.lp_intra_dev,
            self.lp_noise,
            out=np.zeros(self.num_depths),
            where=(self.lp_noise > 1.0),
        )

        intra = self.depth_filter(norm_lp_intra_dev)

        # Inter-frame part

        sf = self.dynamic_sf(self.fast_sf)
        self.fast_lp_mean_sweep = sf * self.fast_lp_mean_sweep + (1.0 - sf) * mean_sweep

        sf = self.dynamic_sf(self.slow_sf)
        self.slow_lp_mean_sweep = sf * self.slow_lp_mean_sweep + (1.0 - sf) * mean_sweep

        inter_diff = self.fast_lp_mean_sweep - self.slow_lp_mean_sweep

        if self.num_removed_pc > 0:
            noise_coeffs = inter_diff @ self.noise_base.transpose()
            inter_diff = np.abs(inter_diff) - np.abs(noise_coeffs @ self.noise_base)
            inter_dev = np.maximum(inter_diff, 0)
        else:
            inter_dev = np.abs(inter_diff)

        sf = self.dynamic_sf(self.inter_dev_sf)
        self.lp_inter_dev = sf * self.lp_inter_dev + (1.0 - sf) * inter_dev

        norm_lp_dev = np.divide(
            self.lp_inter_dev,
            self.lp_noise,
            out=np.zeros_like(self.lp_inter_dev),
            where=(self.lp_noise > 1.0),
        )

        norm_lp_dev *= np.sqrt(self.sweeps_per_frame)

        inter = self.depth_filter(norm_lp_dev)

        # Update the noise base

        if self.num_removed_pc > 0:
            noise_coeffs = self.noise_base @ noise_diff.transpose()
            noise_base_update = noise_coeffs @ noise_diff
            noise_base_update = self.normalize_noise_base(noise_base_update)

            sf = self.dynamic_sf(self.noise_sf)
            self.noise_base = sf * self.noise_base + (1.0 - sf) * noise_base_update
            for i in range(len(self.noise_base)):
                for j in range(i):
                    self.noise_base[i] -= self.noise_base[j] * np.dot(
                        self.noise_base[j], self.noise_base[i]
                    )
                self.noise_base[i] /= np.sqrt(np.sum(np.square(self.noise_base[i])))

        # Detector output

        depthwise_presence = self.inter_weight * inter + self.intra_weight * intra

        max_depthwise_presence = np.max(depthwise_presence)

        sf = self.dynamic_sf(self.output_sf)
        self.presence_score = sf * self.presence_score + (1.0 - sf) * max_depthwise_presence

        presence_detected = self.presence_score > self.threshold

        self.presence_history = np.roll(self.presence_history, -1)
        self.presence_history[-1] = self.presence_score

        if max_depthwise_presence > self.threshold:
            self.presence_distance_index = np.argmax(depthwise_presence)
            self.presence_distance = self.depths[self.presence_distance_index]

        out_data = {
            "frame": frame,
            "fast": self.fast_lp_mean_sweep,
            "slow": self.slow_lp_mean_sweep,
            "noise": self.lp_noise,
            "inter": inter * self.inter_weight,
            "intra": intra * self.intra_weight,
            "depthwise_presence": depthwise_presence,
            "presence_score": self.presence_score,
            "presence_distance_index": self.presence_distance_index,
            "presence_distance": self.presence_distance,
            "presence_history": self.presence_history,
            "presence_detected": presence_detected,
        }

        self.update_index += 1

        return out_data



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import Any, Optional

from PySide6.QtWidgets import QTextEdit

from acconeer.exptool.app.new.app_model import AppModel


class AppModelViewer(QTextEdit):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__()
        app_model.sig_notify.connect(self._update_text)
        self.setFontFamily("monospace")
        self.setMinimumSize(500, 200)
        self.setReadOnly(True)

    @staticmethod
    def _stringify_backend_state(backend_state: Optional[Any]) -> list[str]:
        if backend_state is None:
            return ["None"]

        try:
            members = backend_state.__slots__
        except AttributeError:
            members = backend_state.__dict__

        public_members = [member for member in members if not member.startswith("_")]

        return [
            f">>> {attribute}\n" + f"{getattr(backend_state, attribute)}\n"
            for attribute in public_members
        ]

    def _update_text(self, app_model: AppModel) -> None:
        self.setText(
            "\n".join(
                [
                    "=== AppModel ===",
                    "",
                    f"ConnectionState:        {app_model.connection_state}",
                    f"ConnectionInterface:    {app_model.connection_interface}",
                    f"PluginState:            {app_model.plugin_state}",
                    f"socket_connection_ip:   {app_model.socket_connection_ip}",
                    f"serial_connection_port: {app_model.serial_connection_port}",
                    "",
                    "backend_plugin_state:",
                    *self._stringify_backend_state(app_model.backend_plugin_state),
                ]
            )
        )
